/*
 * $Id: cilk-nd.h,v 1.1.2.6 1997/03/04 00:17:21 fengmd Exp $
 */

/*
 * $Log: cilk-nd.h,v $
 * Revision 1.1.2.6  1997/03/04 00:17:21  fengmd
 * ND_VERBOSE defaults to 0.
 *
 * Revision 1.1.2.5  1997/03/04 00:06:58  fengmd
 * Fix static region to accommodate the region of global initialized variable.
 *
 * Revision 1.1.2.4  1997/03/03 21:02:11  fengmd
 * Add ND_VERBOSE to show the address range of shadow space.
 *
 * Revision 1.1.2.3  1997/02/18 19:56:32  fengmd
 * Add a command-line option for the Nondeterminator's heap size.
 *
 * Revision 1.1.2.2  1997/02/15 02:46:41  fengmd
 * Add Id and Log
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * macros to define the address range of the static region
 */
#if CILK_SOLARIS
extern _begin_Cilk_shared;
extern _end_Cilk_shared;
#define STATIC_ARENA_START      &_begin_Cilk_shared
#define STATIC_ARENA_END        &_end_Cilk_shared

#elif CILK_SOLARIS_THREADS
extern _etext;
extern _end;
#define STATIC_ARENA_START      &_etext
#define STATIC_ARENA_END        &_end

#elif CILK_IRIX_THREADS
extern _fdata;
extern _end;
#define STATIC_ARENA_START      &_fdata
#define STATIC_ARENA_END        &_end

#else
#warning The Nondeterminator has only been tested on Solaris and IRIX
#warning Please define STATIC_ARENA_START and STATIC_ARENA_END in cilk-nd.h
#endif

/*
 * architecture specific macros
 */
#if (CILK_SOLARIS || CILK_SOLARIS_THREADS)
#define WHEN_SOLARIS(x)         x
#define WHEN_IRIX(x)
#define STATIC_AT_LOWER_ADDRESS 1

#elif CILK_IRIX_THREADS
#define WHEN_SOLARIS(x)
#define WHEN_IRIX(x)            x
#define STATIC_AT_LOWER_ADDRESS 0

#else
#define WHEN_SOLARIS(x)
#define WHEN_IRIX(x)    
#define STATIC_AT_LOWER_ADDRESS 1
#endif

/* 
 * macro to display the shadow address range
 */
#define ND_VERBOSE 0

/* 
 * macros to get shadow address of reader and writer
 */
#if (CILK_SOLARIS || CILK_SOLARIS_THREADS)
#define STATIC_READER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) | 0xB3000000))
#define STATIC_WRITER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) | 0xD3000000))
#define HEAP_READER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) & 0xBFFFFFFF))
#define HEAP_WRITER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) & 0xDFFFFFFF))
#elif CILK_IRIX_THREADS
#define STATIC_READER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) + 0x1C000000))
#define STATIC_WRITER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) + 0x3C000000))
#define HEAP_READER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) | 0x20000000))
#define HEAP_WRITER(x) \
  ((DisjointSetMemberT *) (((unsigned int) x) | 0x40000000))
#endif

/* 
 * macro to get shadow address of var and info
 * from shadow address of reader and writer
 */
#define SHADOW_VAR(p)       \
 ((char **) (((unsigned int) p) - ((unsigned int) USE_SHARED_RO(region_size))))
#define SHADOW_INFO(x)      \
 ((DisjointSetMemberT *) (((unsigned int) x) + ((unsigned int) 0x10000000)))
#define SHADOW_INFO_LINE(p) (((unsigned int) *p) >> 8)
#define SHADOW_INFO_FILE(p) (((unsigned int) *p) & 0x000000FF)

/* 
 * macros for determining which region has lower address
 */
#if STATIC_AT_LOWER_ADDRESS
#define STATIC_LOWER(x) x
#define HEAP_LOWER(x)  
#else
#define STATIC_LOWER(x) 
#define HEAP_LOWER(x)   x
#endif

/* 
 * SP-bags constants
 */
#define S_SET        ((DisjointSetMemberT) 0)
#define P_SET        ((DisjointSetMemberT) -1)
#define FIRST_THREAD ((DisjointSetMemberT) 1)

/* 
 * the least significant 28 bits are used for procedure id, the high-order
 * 4 bits are used for storing operator and type 
 */
#define PROC_ID_BIT     28
#define MAX_PROC_NUMBER (1UL<<PROC_ID_BIT)
#define PROC_ID(x)      (((unsigned int) x) & (MAX_PROC_NUMBER - 1))
#define OP_TYPE(x)      \
  ((((unsigned int) x) & (~(MAX_PROC_NUMBER - 1))) >> PROC_ID_BIT)
#define OP_TYPE_PROC_ID(proc_id, op_type) ((op_type << PROC_ID_BIT) + proc_id)
#define DEFAULT_ND_HEAPSIZE (16U * MEGABYTE)

/* 
 * operator takes 2 lower-order bits, and the high-order 2 bits for type 
 */
#define OPERATOR_BIT      2
#define MAX_OPERATOR      (1UL<<OPERATOR_BIT)
#define CHECK_OPERATOR(x) ((x) & (MAX_OPERATOR - 1))
#define TYPE_BIT          2
#define CHECK_TYPE(x)     (((x) & (~(MAX_OPERATOR - 1))) >> OPERATOR_BIT)

/* 
 * operator value, from 0 to 2^OPERATOR_BIT - 1
 */
#define PLUSassign 0
#define DIVassign  1
#define ANDassign  2
#define EQassign   3

/* 
 * type value, from 0 to 2^TYPE_BIT - 1
 */
#define INTtype    0
#define FLOATtype  1
#define POINTtype  2
#define OTHERtype  3

/* 
 * structure to store list of file name 
 */
typedef struct FileListT { 
  struct FileListT *next; 
  char *filename; 
  int index;
} FileListT;

/*
 * variables defined in nondeterminator.c, and used by Cilk rts
 */
DECLARE_SHARED_RO_VARIABLE(extern char*, heap_arena_start, );
DECLARE_SHARED_RO_VARIABLE(extern char*, heap_arena_end, );
DECLARE_PRIVATE_VARIABLE(extern int, lreturn_size, );
DECLARE_SHARED_RO_VARIABLE(extern int, first_stop, );
DECLARE_SHARED_RO_VARIABLE(extern int, float_comm, );
DECLARE_SHARED_RO_VARIABLE(extern int, nd_arena, );

/* 
 * SP-bags action functions, called from cilk2c'ed Cilk program or rts
 */
extern void Cilk_init_spbags(StackFrame *f);
extern void Cilk_lreturn_spbags(StackFrame *f,volatile StackFrame *child);
extern void Cilk_sync_spbags(StackFrame *f);
extern void nondeterminator_init();

/*
 * checking protype, called with gcc on cilk2c'ed Cilk program
 */
#ifdef CILK_NDL

extern void *Cilk_internal_read_check(void *p, size_t size, 
				      int linenum,char *filename,char *name);
extern void *Cilk_internal_read_check4(void *p,
				       int linenum,char *filename,char *name);
extern void *Cilk_internal_read_check8(void *p,
				       int linenum,char *filename,char *name);
extern void *Cilk_internal_write_check(void *p, size_t size, int op_type,
				       int linenum,char *filename,char *name);
extern void *Cilk_internal_write_check4(void *p, int op_type,
					int linenum,char *filename,char *name);
extern void *Cilk_internal_write_check8(void *p, int op_type,
					int linenum,char *filename,char *name);
extern void Cilk_internal_spawnvar_check(void *p, size_t size, int op_type,
			  StackFrame *f,int linenum,char *filename,char *name);

#else

extern void *Cilk_internal_read_check(void *p, size_t size);
extern void *Cilk_internal_read_check4(void *p);
extern void *Cilk_internal_read_check8(void *p);
extern void *Cilk_internal_write_check(void *p, size_t size,int op_type);
extern void *Cilk_internal_write_check4(void *p, int op_type);
extern void *Cilk_internal_write_check8(void *p, int op_type);
extern void Cilk_internal_spawnvar_check(void *p, size_t size,
					 int op_type, StackFrame *f);

#endif

/*
 * checking protype called in Cilk program, transformed in cilk2c.
 * no need to implement.
 */
#ifdef CILK_ND
extern void Cilk_read_check(void *p, size_t size);
extern void Cilk_write_check(void *p, size_t size);
#else
#define Cilk_read_check(p, size)
#define Cilk_write_check(p, size)
#endif

/* 
 * memory check cache 
 */
#define AddressCacheSize 1024
#define CacheAddr(p) ((((unsigned int) (p)) >> 2) & 0x3FF)

typedef struct AddressCacheT {
  void *address;
  size_t size;
  DisjointSetMemberT proc_id;
  unsigned int version;
} AddressCacheT;

/* 
 * sp-bag timing constants
 */
#define HRTIME 575
#define WALL_TIME_TO_MICROSECONDS(x) (0.001*x)
#define ADJUST_TIME(t, c) (t-(((hrtime_t) c)*HRTIME))

/* 
 * sp-bag timing function 
 */
WHEN_CILK_ND_STATS(
void Cilk_spbag_start_timing();
void Cilk_spbag_stop_timing();
void Cilk_spbag_print_timing();
)
