/*
 * $Id: nondeterminator.c,v 1.1.2.6 1997/03/04 00:07:01 fengmd Exp $
 */

/*
 * $Log: nondeterminator.c,v $
 * Revision 1.1.2.6  1997/03/04 00:07:01  fengmd
 * Fix static region to accommodate the region of global initialized variable.
 *
 * Revision 1.1.2.5  1997/03/03 21:02:14  fengmd
 * Add ND_VERBOSE to show the address range of shadow space.
 *
 * Revision 1.1.2.4  1997/02/18 21:10:05  fengmd
 * More message shown when rts error occurs.
 *
 * Revision 1.1.2.3  1997/02/18 19:56:35  fengmd
 * Add a command-line option for the Nondeterminator's heap size.
 *
 * Revision 1.1.2.2  1997/02/15 02:46:43  fengmd
 * Add Id and Log
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <cilk-cilk2c.h>

/***********************************************************\
 * Memory check cache
\***********************************************************/

WHEN_CILK_ND_CACHE(
DECLARE_PRIVATE_VARIABLE(AddressCacheT*, read_cache, );
DECLARE_PRIVATE_VARIABLE(AddressCacheT*, write_cache, );
DECLARE_PRIVATE_VARIABLE(unsigned int, current_version, );
WHEN_CILK_ND_STATS(
DECLARE_PRIVATE_VARIABLE(long long, read_access_cache_hit_count, );
DECLARE_PRIVATE_VARIABLE(long long, write_access_cache_hit_count, );
DECLARE_PRIVATE_VARIABLE(long long, read_access_cache_hit_word_count, );
DECLARE_PRIVATE_VARIABLE(long long, write_access_cache_hit_word_count, );
)
)

WHEN_CILK_ND_CACHE(
static void init_cache()
{
  if ((USE_PRIVATE(read_cache) =
      (AddressCacheT *)malloc(AddressCacheSize * sizeof(AddressCacheT)))
      == NULL)
    Cilk_die("malloc error in get_shadow_space for read_cache\n");

  if ((USE_PRIVATE(write_cache) =
      (AddressCacheT *)malloc(AddressCacheSize * sizeof(AddressCacheT)))
      == NULL)
    Cilk_die("malloc error in get_shadow_space for write_cache\n");
}

inline static int check_address_cache(AddressCacheT *c, void *p, 
   size_t size, DisjointSetMemberT proc_id, unsigned int version)
{
  if ((c->address == p) && (size <= c->size) && (c->proc_id == proc_id) && 
      (c->version == version)) {
    return 1;
  }
  else {
    c->address = p;
    c->size = size;
    c->proc_id = proc_id;
    c->version = version;
    return 0;
  }
}
)


/***********************************************************\
 * region of shared memory
\***********************************************************/

DECLARE_SHARED_RO_VARIABLE(char*, heap_arena_start, );
DECLARE_SHARED_RO_VARIABLE(char*, heap_arena_end, );


/***********************************************************\
 * SP-bags algorithm
\***********************************************************/

#ifdef CILK_ND

DECLARE_PRIVATE_VARIABLE(DisjointSetMemberT*, parent, );
DECLARE_PRIVATE_VARIABLE(DisjointSetMemberT*, rank, );
DECLARE_PRIVATE_VARIABLE(DisjointSetMemberT, next_proc_id, );
DECLARE_PRIVATE_VARIABLE(DisjointSetMemberT, current_proc_id, );
DECLARE_PRIVATE_VARIABLE(DisjointSetMemberT, current_inlet_id, );
DECLARE_PRIVATE_VARIABLE(int, lreturn_size, );
DECLARE_SHARED_RO_VARIABLE(int, first_stop, );
DECLARE_SHARED_RO_VARIABLE(int, float_comm, );
DECLARE_SHARED_RO_VARIABLE(int, nd_arena, );

WHEN_CILK_NDL(
DECLARE_PRIVATE_VARIABLE(FileListT*, filelist, );
DECLARE_PRIVATE_VARIABLE(int, race_linenum, );
DECLARE_PRIVATE_VARIABLE(char*, race_filename, );
DECLARE_PRIVATE_VARIABLE(char*, race_name, );
DECLARE_PRIVATE_VARIABLE(DisjointSetMemberT*, race_party, );
DECLARE_SHARED_RO_VARIABLE(size_t, region_size, );
)

WHEN_CILK_ND_STATS(
DECLARE_PRIVATE_VARIABLE(int, spbags_timing, );

DECLARE_PRIVATE_VARIABLE(Cilk_time, read_access_time, );
DECLARE_PRIVATE_VARIABLE(Cilk_time, write_access_time, );
DECLARE_PRIVATE_VARIABLE(long long, read_access_count, );
DECLARE_PRIVATE_VARIABLE(long long, write_access_count, );
DECLARE_PRIVATE_VARIABLE(long long, read_access_global_count, );
DECLARE_PRIVATE_VARIABLE(long long, write_access_global_count, );
DECLARE_PRIVATE_VARIABLE(long long, read_access_word_count, );
DECLARE_PRIVATE_VARIABLE(long long, write_access_word_count, );

DECLARE_PRIVATE_VARIABLE(Cilk_time, init_time, );
DECLARE_PRIVATE_VARIABLE(int, init_count, );
DECLARE_PRIVATE_VARIABLE(Cilk_time, return_time, );
DECLARE_PRIVATE_VARIABLE(int, return_count, );
DECLARE_PRIVATE_VARIABLE(Cilk_time, sync_time, );
DECLARE_PRIVATE_VARIABLE(int, sync_count, );
)

static void report_race(void *p);
static inline int is_commutative(unsigned int op1, unsigned int op2);
static inline int in_same_set(DisjointSetMemberT p1, DisjointSetMemberT p2);
WHEN_CILK_NDL(
static void save_shadow_info(DisjointSetMemberT *race_party);
static inline char *strip_frame(char *name);
static inline int find_filename(char *filename);
static inline char *get_filename(int index);
)

/* 
 * Disjoint-set operations 
 */
inline static DisjointSetMemberT make_set()
{
  register DisjointSetMemberT i;

  i = USE_PRIVATE(current_proc_id) = USE_PRIVATE(next_proc_id);
  USE_PRIVATE(current_inlet_id) = S_SET;
  WHEN_CILK_ND_CACHE(USE_PRIVATE(current_version) = 0;)
  CILK_CHECK(i < USE_SHARED_RO(nd_arena),
             ("The Nondeterminator runs out %d bytes of disjoint-set space\n"
	      "Please try a larger Nondeterminator heap size (by -nd-heap option)\n",
	      USE_SHARED_RO(nd_arena)));
  USE_PRIVATE(parent)[i] = (DisjointSetMemberT) S_SET;
  USE_PRIVATE(rank)[i] = 0;
  USE_PRIVATE(next_proc_id)++;
  return i;
}

inline static DisjointSetMemberT Union(DisjointSetMemberT x, 
                                       DisjointSetMemberT y)
{
  register DisjointSetMemberT rx, ry;

  rx = USE_PRIVATE(rank)[x];
  ry = USE_PRIVATE(rank)[y];
  if (rx > ry) {
    USE_PRIVATE(parent)[y] = x;
    return x;
  }
  else {
    USE_PRIVATE(parent)[x] = y;
    if (rx == ry)
      USE_PRIVATE(rank)[y]++;
    return y;
  }
}

inline static DisjointSetMemberT find_set(DisjointSetMemberT proc_id)
{
  register DisjointSetMemberT p = proc_id, pp, ppp;

  if (p > S_SET) {
    if ((p = USE_PRIVATE(parent)[p]) > S_SET) {
      if ((p = USE_PRIVATE(parent)[p]) > S_SET) {
	do {
	  pp = p;
	  p = USE_PRIVATE(parent)[p];
	} while (p > S_SET);
	do {
	  ppp = USE_PRIVATE(parent)[proc_id];
	  USE_PRIVATE(parent)[proc_id] = pp;
	  proc_id = ppp;
	} while (proc_id != pp);
      }
    }
  }
  return p;
}

/*
 * initialize the SP-bag per instance of the procedure.
 * Cilk_init_spbags is called in INIT_FRAME when a procedure starts
 */
inline void Cilk_init_spbags(StackFrame *f)
{
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(init_count)++;)

  f->proc_id = f->s_set = make_set();
  f->inlet_id = f->p_set = S_SET;
  WHEN_CILK_ND_CACHE(f->version = 0;)

  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(init_time) += gethrtime() - time_start;)
}

/*
 * union the S-bag of the child into the P-bag of the parent.
 * Cilk_lreturn_spbags is called in _XPOP_FRAME_RESULT or _XPOP_FRAME_NORESULT
 */
inline void Cilk_lreturn_spbags(StackFrame *f, volatile StackFrame *child)
{
  register int more_size;
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(return_count)++;)

  USE_PRIVATE(parent)[f->p_set = Union(f->p_set, child->s_set)] = P_SET;
  USE_PRIVATE(current_proc_id) = f->proc_id;
  USE_PRIVATE(current_inlet_id) = f->inlet_id;
  WHEN_CILK_ND_CACHE(USE_PRIVATE(current_version) = (++(f->version));)

  /* clean up shadow space of the child frame */
  more_size = ((USE_PRIVATE(lreturn_size) + sizeof(int) - 1) & 
	       (~(sizeof(int) -1))) - sizeof(StackFrame);
  memset(HEAP_READER(child) + sizeof(StackFrame)/sizeof(DisjointSetMemberT),
	 S_SET, more_size);
  memset(HEAP_WRITER(child) + sizeof(StackFrame)/sizeof(DisjointSetMemberT),
	 S_SET, more_size);

  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(return_time) += gethrtime() - time_start;)
}

/*
 * union the P-bag of the current procedure into its S-bag.
 * Cilk_sync_spbags is called in _AT_SYNC_FAST
 */
inline void Cilk_sync_spbags(StackFrame *f)
{
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(sync_count)++;)

  USE_PRIVATE(parent)[f->s_set = Union(f->p_set, f->s_set)] = S_SET;
  f->inlet_id = f->p_set = S_SET; /* after sync, a new inlet id is needed */

  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(sync_time) += gethrtime() - time_start;)
}

/*
 * routine to check a read access of an address p with p's reader.
 * p's reader may be updated.
 */
inline static int read_check_reader(DisjointSetMemberT *readerp,
				    DisjointSetMemberT my_proc_id)
{
  register DisjointSetMemberT reader = *readerp;

  if (find_set(reader) == S_SET) {
    *readerp = my_proc_id;
    WHEN_CILK_NDL(save_shadow_info(readerp);)
    return 0;
  }
  return 1;
}

/*
 * routine to check a read access of an address p with p's writer.
 */
inline static void read_check_writer(void *p, DisjointSetMemberT *writerp)
{
  if (find_set(PROC_ID(*writerp)) == P_SET) {
    WHEN_CILK_NDL(USE_PRIVATE(race_party) = writerp;)
    report_race(p);
  }
}

/*
 * routine to check a write access of an address p with p's reader.
 * p's writer and reader may also be updated.
 */
inline static void write_check(void *p, DisjointSetMemberT *readerp,
			       DisjointSetMemberT *writerp,
			       DisjointSetMemberT my_proc_id, int op_type)
{
  register DisjointSetMemberT reader = *readerp;

  if (find_set(reader) == P_SET) {
    register DisjointSetMemberT writer = *writerp;

    if (!((PROC_ID(writer) == USE_PRIVATE(current_inlet_id)) &&
	  is_commutative(OP_TYPE(writer), op_type))) {
      WHEN_CILK_NDL(USE_PRIVATE(race_party) = readerp;)
      report_race(p);
    }
  }
  else {
    *readerp = my_proc_id;
    *writerp = OP_TYPE_PROC_ID(my_proc_id, op_type);
    WHEN_CILK_NDL(save_shadow_info(readerp);)
    WHEN_CILK_NDL(save_shadow_info(writerp);)
  }
}

/*
 * check the read access in region [p, p+size)
 */
#ifdef CILK_NDL
void *Cilk_internal_read_check(void *p, size_t size,
			       int linenum, char *filename, char *name)
#else
void *Cilk_internal_read_check(void *p, size_t size)
#endif
{
  register int *pp;
  register DisjointSetMemberT my_proc_id = USE_PRIVATE(current_proc_id);
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_count)++;)

  WHEN_CILK_ND_CACHE(
    if (check_address_cache(&(USE_PRIVATE(read_cache)[CacheAddr(p)]), p, size, 
			    my_proc_id, USE_PRIVATE(current_version))) {
      WHEN_CILK_ND_STATS_DO(
        USE_PRIVATE(read_access_cache_hit_count)++;
        USE_PRIVATE(read_access_cache_hit_word_count) += 
	            (size + sizeof(int) - 1) / sizeof(int);
	USE_PRIVATE(read_access_time) += gethrtime() - time_start;
      )
      return p;
    }
  )

  WHEN_CILK_NDL(USE_PRIVATE(race_linenum) = linenum; 
		USE_PRIVATE(race_filename) = filename;
		USE_PRIVATE(race_name) = name;)

  size = (size + sizeof(int) - 1) & (~(sizeof(int) -1));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp < (int *) USE_SHARED_RO(heap_arena_end)) {
      for (; size; size -= sizeof(int), pp++) {
	if (read_check_reader(HEAP_READER(pp), my_proc_id))
	  read_check_writer(pp, HEAP_WRITER(pp));

	WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_word_count)++;)
      }
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_time) += 
				   gethrtime() - time_start;)
    }
  STATIC_LOWER(})
  else if (pp >= (int *) STATIC_ARENA_START) {
    if (pp < (int *) STATIC_ARENA_END) {
      for (; size; size -= sizeof(int), pp++) {
	if (read_check_reader(STATIC_READER(pp), my_proc_id))
	  read_check_writer(pp, STATIC_WRITER(pp));

	WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_word_count)++;)
      }
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_time) += 
				   gethrtime() - time_start;)
    }
  }
  HEAP_LOWER(})
  
  return p;
}

/*
 * check the read access in region [p, p+4).
 * this is to optimize the common case of "read an integer or float".
 */
#ifdef CILK_NDL
void *Cilk_internal_read_check4(void *p, int linenum, char *filename,char *name)
#else
void *Cilk_internal_read_check4(void *p)
#endif
{
  register DisjointSetMemberT my_proc_id = USE_PRIVATE(current_proc_id);
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_count)++;)

  WHEN_CILK_ND_CACHE(
    if (check_address_cache(&(USE_PRIVATE(read_cache)[CacheAddr(p)]), p, 4, 
			    my_proc_id, USE_PRIVATE(current_version))) {
      WHEN_CILK_ND_STATS_DO(
	USE_PRIVATE(read_access_cache_hit_count)++;
        USE_PRIVATE(read_access_cache_hit_word_count)++;
	USE_PRIVATE(read_access_time) += gethrtime() - time_start;
      )
      return p;
    }
  )

  WHEN_CILK_NDL(USE_PRIVATE(race_linenum) = linenum; 
		USE_PRIVATE(race_filename) = filename;
		USE_PRIVATE(race_name) = name;)

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p < (void *) USE_SHARED_RO(heap_arena_end)) {
      if (read_check_reader(HEAP_READER(p), my_proc_id))
	read_check_writer(p, HEAP_WRITER(p));

      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_word_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_time) += 
				   gethrtime() - time_start;)
    }
  STATIC_LOWER(})
  else if (p >= (void *) STATIC_ARENA_START) {
    if (p < (void *) STATIC_ARENA_END) {
      if (read_check_reader(STATIC_READER(p), my_proc_id))
	read_check_writer(p, STATIC_WRITER(p));

      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_word_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_time) += 
				   gethrtime() - time_start;)
    }
  }
  HEAP_LOWER(})

  return p;
}

/*
 * check the read access in region [p, p+8).
 * this is to optimize the common case of "read a double".
 */
#ifdef CILK_NDL
void *Cilk_internal_read_check8(void *p, int linenum, char *filename,char *name)
#else
void *Cilk_internal_read_check8(void *p)
#endif
{
  register *pp;
  register DisjointSetMemberT my_proc_id = USE_PRIVATE(current_proc_id);
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_count)++;)

  WHEN_CILK_ND_CACHE(
    if (check_address_cache(&(USE_PRIVATE(read_cache)[CacheAddr(p)]), p, 8, 
			    my_proc_id, USE_PRIVATE(current_version))) {
      WHEN_CILK_ND_STATS_DO(
	USE_PRIVATE(read_access_cache_hit_count)++;
        USE_PRIVATE(read_access_cache_hit_word_count)+=2;
	USE_PRIVATE(read_access_time) += gethrtime() - time_start;
      )
      return p;
    }
  )

  WHEN_CILK_NDL(USE_PRIVATE(race_linenum) = linenum; 
		USE_PRIVATE(race_filename) = filename;
		USE_PRIVATE(race_name) = name;)

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p < (void *) USE_SHARED_RO(heap_arena_end)) {
      if (read_check_reader(HEAP_READER(p), my_proc_id))
	read_check_writer(p, HEAP_WRITER(p));
      
      pp = p + 4;
      if (read_check_reader(HEAP_READER(pp), my_proc_id))
	read_check_writer(pp, HEAP_WRITER(pp));
      
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_word_count)+=2;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_time) += 
				   gethrtime() - time_start;)
    }
  STATIC_LOWER(})
  else if (p >= (void *) STATIC_ARENA_START) {
    if (p < (void *) STATIC_ARENA_END) {
      if (read_check_reader(STATIC_READER(p), my_proc_id))
	read_check_writer(p, STATIC_WRITER(p));
      
      pp = p + 4;
      if (read_check_reader(STATIC_READER(pp), my_proc_id))
	read_check_writer(pp, STATIC_WRITER(pp));
      
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_word_count)+=2;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(read_access_time) += 
				   gethrtime() - time_start;)
    }
  }
  HEAP_LOWER(})

  return p;
}

/*
 * check the write access in region [p, p+size)
 */
#ifdef CILK_NDL
void *Cilk_internal_write_check(void *p, size_t size, int op_type, 
				int linenum, char *filename, char *name)
#else
void *Cilk_internal_write_check(void *p, size_t size, int op_type)
#endif
{
  register int *pp;
  register DisjointSetMemberT my_proc_id = USE_PRIVATE(current_proc_id);
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_count)++;)

  WHEN_CILK_NDL(USE_PRIVATE(race_linenum) = linenum; 
		USE_PRIVATE(race_filename) = filename;
		USE_PRIVATE(race_name) = name;)

  WHEN_CILK_ND_CACHE(
    if (check_address_cache(&(USE_PRIVATE(write_cache)[CacheAddr(p)]), p,size, 
			    OP_TYPE_PROC_ID(my_proc_id, op_type),
			    USE_PRIVATE(current_version))) {
      WHEN_CILK_ND_STATS_DO(
	USE_PRIVATE(write_access_cache_hit_count)++;
        USE_PRIVATE(write_access_cache_hit_word_count) += 
	            (size + sizeof(int) - 1) / sizeof(int);
	USE_PRIVATE(write_access_time) += gethrtime() - time_start;
      )
      return p;
    }
  )

  size = (size + sizeof(int) - 1) & (~(sizeof(int) -1));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp < (int *) USE_SHARED_RO(heap_arena_end)) {
      for (; size; size -= sizeof(int), pp++) {
	write_check(pp, HEAP_READER(pp), HEAP_WRITER(pp), my_proc_id, op_type);

	WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_word_count)++;)
      }
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_time) += 
				   gethrtime() - time_start;)
    }
  STATIC_LOWER(})
  else if (pp >= (int *) STATIC_ARENA_START) {
    if (pp < (int *) STATIC_ARENA_END) {
      for (; size; size -= sizeof(int), pp++) {
	write_check(pp, STATIC_READER(pp), STATIC_WRITER(pp), my_proc_id, op_type);

	WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_word_count)++;)
      }
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_time) += 
				   gethrtime() - time_start;)
    }
  }
  HEAP_LOWER(})

  return p;
}

/*
 * check the write access in region [p, p+4).
 * this is to optimize the common case of "write an integer or float".
 */
#ifdef CILK_NDL
void *Cilk_internal_write_check4(void *p, int op_type, int linenum, 
				 char *filename, char *name)
#else
void *Cilk_internal_write_check4(void *p, int op_type)
#endif
{
  register DisjointSetMemberT my_proc_id = USE_PRIVATE(current_proc_id);
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_count)++;)

  WHEN_CILK_NDL(USE_PRIVATE(race_linenum) = linenum; 
		USE_PRIVATE(race_filename) = filename;
		USE_PRIVATE(race_name) = name;)

  WHEN_CILK_ND_CACHE(
    if (check_address_cache(&(USE_PRIVATE(write_cache)[CacheAddr(p)]), p, 4, 
			    OP_TYPE_PROC_ID(my_proc_id, op_type),
			    USE_PRIVATE(current_version))) {
      WHEN_CILK_ND_STATS_DO(
	USE_PRIVATE(write_access_cache_hit_count)++;
        USE_PRIVATE(write_access_cache_hit_word_count)++;
	USE_PRIVATE(write_access_time) += gethrtime() - time_start;
      )
      return p;
    }
  )

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p < (void *) USE_SHARED_RO(heap_arena_end)) {
      write_check(p, HEAP_READER(p), HEAP_WRITER(p), my_proc_id, op_type);

      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_word_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_time) += 
				   gethrtime() - time_start;)
    }
  STATIC_LOWER(})
  else if (p >= (void *) STATIC_ARENA_START) {
    if (p < (void *) STATIC_ARENA_END) {
      write_check(p, STATIC_READER(p), STATIC_WRITER(p), my_proc_id, op_type);

      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_word_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_time) += 
				   gethrtime() - time_start;)
    }
  }
  HEAP_LOWER(})

  return p;
}

/*
 * check the write access in region [p, p+8).
 * this is to optimize the common case of "write a double".
 */
#ifdef CILK_NDL
void *Cilk_internal_write_check8(void *p, int op_type, int linenum, 
				 char *filename, char *name)
#else
void *Cilk_internal_write_check8(void *p, int op_type)
#endif
{
  register void *pp;
  register DisjointSetMemberT my_proc_id = USE_PRIVATE(current_proc_id);
  WHEN_CILK_ND_STATS(Cilk_time time_start = 0;)

  WHEN_CILK_ND_STATS_DO(time_start = gethrtime();)
  WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_count)++;)

  WHEN_CILK_NDL(USE_PRIVATE(race_linenum) = linenum; 
		USE_PRIVATE(race_filename) = filename;
		USE_PRIVATE(race_name) = name;)

  WHEN_CILK_ND_CACHE(
    if (check_address_cache(&(USE_PRIVATE(write_cache)[CacheAddr(p)]), p, 8, 
			    OP_TYPE_PROC_ID(my_proc_id, op_type),
			    USE_PRIVATE(current_version))) {
      WHEN_CILK_ND_STATS_DO(
	USE_PRIVATE(write_access_cache_hit_count)++;
        USE_PRIVATE(write_access_cache_hit_word_count)+=2; 
	USE_PRIVATE(write_access_time) += gethrtime() - time_start;
      )
      return p;
    }
  )

  if (p >= (void *) USE_SHARED_RO(heap_arena_start)) {
    if (p < (void *) USE_SHARED_RO(heap_arena_end)) {
      write_check(p, HEAP_READER(p), HEAP_WRITER(p), my_proc_id, op_type);

      pp = p + 4;
      write_check(pp, HEAP_READER(pp), HEAP_WRITER(pp), my_proc_id, op_type);

      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_word_count)+=2;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_time) += 
				   gethrtime() - time_start;)
    }
  STATIC_LOWER(})
  else if (p >= (void *) STATIC_ARENA_START) {
    if (p < (void *) STATIC_ARENA_END) {
      write_check(p, STATIC_READER(p), STATIC_WRITER(p), my_proc_id, op_type);
      
      pp = p + 4;
      write_check(pp, STATIC_READER(pp), STATIC_WRITER(pp), my_proc_id, op_type);

      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_word_count)+=2;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_global_count)++;)
      WHEN_CILK_ND_STATS_DO(USE_PRIVATE(write_access_time) += 
				   gethrtime() - time_start;)
    }
  }
  HEAP_LOWER(})

  return p;
}

inline static DisjointSetMemberT make_inlet_id(StackFrame *f)
{
  register DisjointSetMemberT i;

  i = f->inlet_id = USE_PRIVATE(next_proc_id);
  CILK_CHECK(i < USE_SHARED_RO(nd_arena),
             ("The Nondeterminator runs out %d bytes of disjoint-set space\n"
	      "Please try a larger Nondeterminator heap size (by -nd-heap option)\n",
	       USE_SHARED_RO(nd_arena)));
  USE_PRIVATE(rank)[i] = 0;
  USE_PRIVATE(parent)[f->p_set = Union(f->p_set, i)] = P_SET;
  USE_PRIVATE(next_proc_id)++;
  return i;
}

inline void spawnvar_check(void *p, DisjointSetMemberT *readerp,
			   DisjointSetMemberT *writerp,
			   DisjointSetMemberT my_proc_id, int op_type,
			   StackFrame *f)
{
  register DisjointSetMemberT reader = *readerp;
  register DisjointSetMemberT writer = *writerp;

  if (find_set(reader) == P_SET) {
    if (!((PROC_ID(writer) == f->inlet_id) &&
          is_commutative(OP_TYPE(writer), op_type))) {
      WHEN_CILK_NDL(USE_PRIVATE(race_party) = readerp;)
      report_race(p);
    }
  }
  else {
    if (!in_same_set(reader, my_proc_id)) {
      register DisjointSetMemberT inlet_id;

      inlet_id = (f->inlet_id == S_SET) ? make_inlet_id(f) : f->inlet_id;
      *readerp = inlet_id;
      *writerp = OP_TYPE_PROC_ID(inlet_id, op_type);
      WHEN_CILK_NDL(save_shadow_info(readerp);)
      WHEN_CILK_NDL(save_shadow_info(writerp);)
    }
    else if (!in_same_set(PROC_ID(writer), my_proc_id)) {
      register DisjointSetMemberT inlet_id;

      inlet_id = (f->inlet_id == S_SET) ? make_inlet_id(f) : f->inlet_id;
      *writerp = OP_TYPE_PROC_ID(inlet_id, op_type);
      WHEN_CILK_NDL(save_shadow_info(writerp);)
    }
  }
}

/*
 * check commutative of op_type in the shadow space
 */
#ifdef CILK_NDL
void Cilk_internal_spawnvar_check(void *p, size_t size, int op_type,
		        StackFrame *f, int linenum, char *filename, char *name)
#else
void Cilk_internal_spawnvar_check(void *p, size_t size, int op_type,
				  StackFrame *f)
#endif
{
  register int *pp;
  register DisjointSetMemberT my_proc_id = USE_PRIVATE(current_proc_id);

  WHEN_CILK_NDL(USE_PRIVATE(race_linenum) = linenum; 
		USE_PRIVATE(race_filename) = filename;
		USE_PRIVATE(race_name) = name;)

  size = (size + sizeof(int) - 1) & (~(sizeof(int) -1));
  pp = (int *) (((unsigned int) p) & (~(sizeof(int) -1)));

  if (pp >= (int *) USE_SHARED_RO(heap_arena_start)) {
    if (pp < (int *) USE_SHARED_RO(heap_arena_end)) {
      for (; size; size -= sizeof(int), pp++) {
        spawnvar_check(pp, HEAP_READER(pp), HEAP_WRITER(pp), my_proc_id, 
		       op_type, f);
      }
    }
  STATIC_LOWER(})
  else if (pp >= (int *) STATIC_ARENA_START) {
    if (pp < (int *) STATIC_ARENA_END) {
      for (; size; size -= sizeof(int), pp++) {
        spawnvar_check(pp, STATIC_READER(pp), STATIC_WRITER(pp), my_proc_id, 
		       op_type, f);
      }
    }
  }
  else
    CILK_CHECK(pp, ("spawn receiver address %p out of range\n", pp));
  HEAP_LOWER(})
}

/*
 * report race
 */
static void report_race(void *p)
{
  fprintf(stderr, "Determinacy race @ 0x%p", p);
  WHEN_CILK_NDL(
    fprintf(stderr, ": `%s'(%s:%d) with `%s'(%s:%d)",
      strip_frame(*(SHADOW_VAR(SHADOW_INFO(USE_PRIVATE(race_party))))),
      get_filename(SHADOW_INFO_FILE(SHADOW_INFO(USE_PRIVATE(race_party)))),
      SHADOW_INFO_LINE(SHADOW_INFO(USE_PRIVATE(race_party))),
      strip_frame(USE_PRIVATE(race_name)),
      USE_PRIVATE(race_filename), USE_PRIVATE(race_linenum));
  )
  fprintf(stderr, "\n");
  if (USE_SHARED_RO(first_stop)) 
    Cilk_die("The first determinacy race is found, and the program is terminated\n");
}

/* 
 * test whether two operations are communitive or not 
 */
static inline int is_commutative(unsigned int op1, unsigned int op2)
{
  register unsigned int operator1, operator2, type1, type2;

  operator1 = CHECK_OPERATOR(op1);
  operator2 = CHECK_OPERATOR(op2);
  type1 = CHECK_TYPE(op1);
  type2 = CHECK_TYPE(op2);

  CILK_CHECK(type1 == type2, ("type mismatch in is_commutative\n"));
  if (operator1 == operator2) {
    switch (type1) {
    case INTtype:
      return ((operator1 <= ANDassign) && (operator1 != DIVassign)) ? 1 : 0;
    case FLOATtype:
      return USE_SHARED_RO(float_comm) ? 
               ((operator1 <= DIVassign) ? 1 : 0) : 0;
    case POINTtype:
      return (operator1 <= PLUSassign) ? 1 : 0;
    case OTHERtype:
      return 0;
    default:
      CILK_CHECK(type1 > OTHERtype, ("wrong type in is_commutative\n"));
      return 0;
    }
  }
  else
    return 0;
}

/*
 * test whether two procedures are in the same set
 */
static inline int in_same_set(DisjointSetMemberT p1, DisjointSetMemberT p2)
{
  register DisjointSetMemberT pp1, pp2;

  if (p1 > S_SET) {
    for (pp1 = USE_PRIVATE(parent)[p1]; 
	 pp1 > S_SET; 
	 p1 = pp1, pp1 = USE_PRIVATE(parent)[pp1]);
  }
  if (p2 > S_SET) {
    for (pp2 = USE_PRIVATE(parent)[p2]; 
	 pp2 > S_SET; 
	 p2 = pp2, pp2 = USE_PRIVATE(parent)[pp2]);
  }
  return (p1 == p2);
}

/* initialize private variables */
static void init_private()
{
  USE_PRIVATE(next_proc_id) = FIRST_THREAD;

  WHEN_CILK_NDL(USE_PRIVATE(filelist) = NULL;)

  WHEN_CILK_ND_STATS(
  USE_PRIVATE(spbags_timing) = 0;
  USE_PRIVATE(read_access_time) = 0;
  USE_PRIVATE(write_access_time) = 0;
  USE_PRIVATE(read_access_count) = 0;
  USE_PRIVATE(write_access_count) = 0;
  USE_PRIVATE(read_access_global_count) = 0;
  USE_PRIVATE(write_access_global_count) = 0;
  USE_PRIVATE(read_access_word_count) = 0;
  USE_PRIVATE(write_access_word_count) = 0;
  USE_PRIVATE(init_time) = 0;
  USE_PRIVATE(init_count) = 0;
  USE_PRIVATE(return_time) = 0;
  USE_PRIVATE(return_count) = 0;
  USE_PRIVATE(sync_time) = 0;
  USE_PRIVATE(sync_count) = 0;

  WHEN_CILK_ND_CACHE(
  USE_PRIVATE(read_access_cache_hit_count) = 0;
  USE_PRIVATE(write_access_cache_hit_count) = 0;
  USE_PRIVATE(read_access_cache_hit_word_count) = 0;
  USE_PRIVATE(write_access_cache_hit_word_count) = 0;
  )
  )
}

/* malloc space for parent and rank used in disjoint-set */
static void init_disjoint_set()
{
  CILK_CHECK(USE_SHARED_RO(nd_arena) <= MAX_PROC_NUMBER,
	     ("the Nondeterminator acquires heap size larger than the %d bytes limit\n"
	      "Please run with smaller Nondeterminator heap size (by -nd-heap option)\n",
	      MAX_PROC_NUMBER));
  if ((USE_PRIVATE(parent) =
      (DisjointSetMemberT *)malloc(USE_SHARED_RO(nd_arena)*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc error in init_disjoint_set for parent\n"
	     "Please run with smaller Nondeterminator heap size (by -nd-heap option)\n");

  if ((USE_PRIVATE(rank) = 
      (DisjointSetMemberT *)malloc(USE_SHARED_RO(nd_arena)*
				   sizeof(DisjointSetMemberT))) == NULL)
    Cilk_die("malloc error in init_disjoint_set for rank\n"
	     "Please run with smaller Nondeterminator heap size (by -nd-heap option)\n");
}

/* set up the reader and writer shadow space */
static void init_shadow_space()
{
  void *p;
  size_t static_size, heap_size;
  char *static_reader_fname = "/tmp/_Cilk_static_reader";
  char *static_writer_fname = "/tmp/_Cilk_static_writer";
  char *heap_reader_fname = "/tmp/_Cilk_heap_reader";
  char *heap_writer_fname = "/tmp/_Cilk_heap_writer";
  int static_reader_fd, static_writer_fd;
  int heap_reader_fd, heap_writer_fd;
  void *start_static_read, *start_static_write;
  void *start_heap_read, *start_heap_write;
  size_t start_static_arena, start_heap_arena, page_size;
  int junk = 42;
  WHEN_CILK_NDL(
  char *static_reader_info_fname = "/tmp/_Cilk_static_reader_info";
  char *static_writer_info_fname = "/tmp/_Cilk_static_writer_info";
  char *heap_reader_info_fname = "/tmp/_Cilk_heap_reader_info";
  char *heap_writer_info_fname = "/tmp/_Cilk_heap_writer_info";
  int static_reader_info_fd;
  int static_writer_info_fd;
  int heap_reader_info_fd;
  int heap_writer_info_fd;
  )

  /* align the starting address of static and heap regions */
  page_size = Cilk_get_page_size();
  start_static_arena = ((size_t) STATIC_ARENA_START) / page_size * page_size;
  start_heap_arena = ((size_t) USE_SHARED_RO(heap_arena_start)) /
                     page_size * page_size;

  /* get the size of static and heap regions */
  static_size = (size_t) STATIC_ARENA_END - start_static_arena;
  heap_size = (size_t) USE_SHARED_RO(heap_arena_end) - start_heap_arena;
  WHEN_CILK_NDL(INIT_SHARED_RO(region_size, static_size > heap_size ? 
			                      static_size : heap_size);)

  /* get starting addresses of shadow spaces */
  start_static_read = STATIC_READER(start_static_arena);
  start_static_write = STATIC_WRITER(start_static_arena);
  start_heap_read = HEAP_READER(start_heap_arena);
  start_heap_write = HEAP_WRITER(start_heap_arena);

#if ND_VERBOSE
  printf("static 0x%p (align to 0x%x) --- 0x%p\n", 
	 STATIC_ARENA_START, start_static_arena, STATIC_ARENA_END);
  printf("    reader 0x%p", start_static_read);
  WHEN_CILK_NDL(printf(", reader info 0x%p, reader var 0x%p",
		SHADOW_INFO(start_static_read), 
		SHADOW_VAR(SHADOW_INFO(start_static_read)));)
  printf("\n    writer 0x%p", start_static_write);
  WHEN_CILK_NDL(printf(", writer info 0x%p, writer var 0x%p",
		SHADOW_INFO(start_static_write),
		SHADOW_VAR(SHADOW_INFO(start_static_write)));)
  printf("\nheap 0x%p (align to 0x%x) --- 0x%p\n",
	 USE_SHARED_RO(heap_arena_start), start_heap_arena,
	 USE_SHARED_RO(heap_arena_end));
  printf("    reader 0x%p", start_heap_read);
  WHEN_CILK_NDL(printf(", reader info 0x%p, reader var 0x%p",
		SHADOW_INFO(start_heap_read), 
		SHADOW_VAR(SHADOW_INFO(start_heap_read)));)
  printf("\n    writer 0x%p", start_heap_write);
  WHEN_CILK_NDL(printf(", writer info 0x%p, writer var 0x%p",
		SHADOW_INFO(start_heap_write),
		SHADOW_VAR(SHADOW_INFO(start_heap_write)));)
  printf("\n");
#endif	       

  /* open files for mmap */
  static_reader_fd = open(static_reader_fname, O_RDWR | O_CREAT, 0600);
  static_writer_fd = open(static_writer_fname, O_RDWR | O_CREAT, 0600);
  heap_reader_fd = open(heap_reader_fname, O_RDWR | O_CREAT, 0600);
  heap_writer_fd = open(heap_writer_fname, O_RDWR | O_CREAT, 0600);
  WHEN_CILK_NDL(
  static_reader_info_fd = open(static_reader_info_fname, O_RDWR|O_CREAT, 0600);
  static_writer_info_fd = open(static_writer_info_fname, O_RDWR|O_CREAT, 0600);
  heap_reader_info_fd = open(heap_reader_info_fname, O_RDWR | O_CREAT, 0600);
  heap_writer_info_fd = open(heap_writer_info_fname, O_RDWR | O_CREAT, 0600);
  )

  CILK_CHECK(static_reader_fd >= 0, ("Open %s error\n", static_reader_fname));
  CILK_CHECK(static_writer_fd >= 0, ("Open %s error\n", static_writer_fname));
  CILK_CHECK(heap_reader_fd >= 0, ("Open %s error\n", heap_reader_fname));
  CILK_CHECK(heap_writer_fd >= 0, ("Open %s error\n", heap_writer_fname));
  WHEN_CILK_NDL(
  CILK_CHECK(static_reader_info_fd >= 0, 
	     ("Open %s error\n", static_reader_info_fname));
  CILK_CHECK(static_writer_info_fd >= 0, 
	     ("Open %s error\n", static_writer_info_fname));
  CILK_CHECK(heap_reader_info_fd >= 0, 
	     ("Open %s error\n", heap_reader_info_fname));
  CILK_CHECK(heap_writer_info_fd >= 0, 
	     ("Open %s error\n, heap_writer_info_fname"));
  )

  /* delete temporary file */
  unlink(static_reader_fname);
  unlink(static_writer_fname);
  unlink(heap_reader_fname);
  unlink(heap_writer_fname);
  WHEN_CILK_NDL(
  unlink(static_reader_info_fname);
  unlink(static_writer_info_fname);
  unlink(heap_reader_info_fname);
  unlink(heap_writer_info_fname);
  )

  CILK_CHECK(lseek(static_reader_fd, 
		   static_size - sizeof(int), SEEK_SET) ==
	     static_size - sizeof(int),
	     ("error in init_shadow_space to lseek static_reader_fd!\n"));
  CILK_CHECK(write(static_reader_fd, &junk, sizeof(int)) == sizeof(int),
	     ("error in init_shadow_space to write static_reader_fd!\n"));

  CILK_CHECK(lseek(static_writer_fd, 
		   static_size - sizeof(int), SEEK_SET) ==
	     static_size - sizeof(int),
	     ("error in init_shadow_space to lseek static_writer_fd!\n"));
  CILK_CHECK(write(static_writer_fd, &junk, sizeof(int)) == sizeof(int),
	     ("error in init_shadow_space to write static_writer_fd!\n"));

  CILK_CHECK(lseek(heap_reader_fd, 
		   heap_size - sizeof(int), SEEK_SET) ==
	     heap_size - sizeof(int),
	     ("error in init_shadow_space to lseek heap_reader_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  CILK_CHECK(write(heap_reader_fd, &junk, sizeof(int)) == sizeof(int),
             ("error in init_shadow_space to write heap_reader_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));

  CILK_CHECK(lseek(heap_writer_fd, 
		   heap_size - sizeof(int), SEEK_SET) ==
	     heap_size - sizeof(int),
	     ("error in init_shadow_space to lseek heap_writer_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  CILK_CHECK(write(heap_writer_fd, &junk, sizeof(int)) == sizeof(int),
	     ("error in init_shadow_space to write heap_writer_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));

  WHEN_CILK_NDL(
  CILK_CHECK(lseek(static_reader_info_fd, 
		   static_size + heap_size - sizeof(int), SEEK_SET) ==
	     static_size + heap_size - sizeof(int),
	     ("error in init_shadow_space to lseek static_reader_info_fd!\n"));
  CILK_CHECK(write(static_reader_info_fd, &junk, sizeof(int)) == sizeof(int),
	     ("error in init_shadow_space to write static_reader_info_fd!\n"));

  CILK_CHECK(lseek(static_writer_info_fd, 
		   static_size + heap_size - sizeof(int), SEEK_SET) ==
	     static_size + heap_size - sizeof(int),
	     ("error in init_shadow_space to lseek static_writer_info_fd!\n"));
  CILK_CHECK(write(static_writer_info_fd, &junk, sizeof(int)) == sizeof(int),
	     ("error in init_shadow_space to write static_writer_info_fd!\n"));

  CILK_CHECK(lseek(heap_reader_info_fd, 
		   2*heap_size - sizeof(int), SEEK_SET) ==
	     2*heap_size - sizeof(int),
	     ("error in init_shadow_space to lseek heap_reader_info_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  CILK_CHECK(write(heap_reader_info_fd, &junk, sizeof(int)) == sizeof(int),
             ("error in init_shadow_space to write heap_reader_info_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));

  CILK_CHECK(lseek(heap_writer_info_fd, 
		   2*heap_size - sizeof(int), SEEK_SET) ==
	     2*heap_size - sizeof(int),
	     ("error in init_shadow_space to lseek heap_writer_info_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  CILK_CHECK(write(heap_writer_info_fd, &junk, sizeof(int)) == sizeof(int),
	     ("error in init_shadow_space to write heap_writer_info_fd!\n"
	      "Please run with smaller heap size (by -heap option)\n"));
  )

  /* make sure the shadow space not overlap */
  CILK_CHECK(((unsigned int) USE_SHARED_RO(heap_arena_start) & 0xF0000000) == 
	     ((unsigned int) USE_SHARED_RO(heap_arena_end) &0xF0000000),
	     ("heap address [0x%p, 0x%p] should be within 1M range\n"
	      "Please run with smaller heap size (by -heap option)\n",
	      USE_SHARED_RO(heap_arena_start), 
	      USE_SHARED_RO(heap_arena_end)));
  WHEN_SOLARIS(
  CILK_CHECK(((unsigned int) USE_SHARED_RO(heap_arena_start) & 0xF0000000) == 
	     0xE0000000,
	     ("heap address (0x%p) should start from 0xe_______\n", 
	      USE_SHARED_RO(heap_arena_start)));)
  WHEN_IRIX(
  CILK_CHECK(((unsigned int) USE_SHARED_RO(heap_arena_start) & 0xF0000000) == 
	     0x00000000,
	     ("heap address (0x%p) should start from 0x0_______\n", 
	      USE_SHARED_RO(heap_arena_start)));)
  CILK_CHECK((char *) start_heap_read + heap_size <= 
	     (char *) start_static_read,
	     ("Overlap between heap reader with static reader\n"
	      "Please change the shadow address in cilk-nd.h\n"
	      "See the Cilk manual for more explanation\n"));
  WHEN_CILK_NDL(
  CILK_CHECK((char *) start_static_read + static_size <= 
	     (char *) SHADOW_VAR(SHADOW_INFO(start_heap_read)),
	     ("Overlap between static reader with heap reader info\n"
	      "Please change the shadow address and info address in cilk-nd.h\n"
	      "See the Cilk manual for more explanation\n"));
  CILK_CHECK((char *) SHADOW_INFO(start_heap_read) + heap_size <=
	      (char *) SHADOW_VAR(SHADOW_INFO(start_static_read)),
	     ("Overlap between heap reader info with static reader info\n"
	      "Please change the shadow address and info address in cilk-nd.h\n"
	      "See the Cilk manual for more explanation\n"));
  CILK_CHECK((char *) SHADOW_INFO(start_static_read) + static_size <= 
	      (char *) start_heap_write,
	     ("Overlap between static reader info with heap writer\n"
	      "Please change the shadow address and info address in cilk-nd.h\n"
	      "See the Cilk manual for more explanation\n"));
  )
  CILK_CHECK((char *) start_heap_write + heap_size <= 
	      (char *) start_static_write,
	     ("Overlap between heap writer info with static writer\n"
	      "Please change the shadow address and in cilk-nd.h\n"
	      "See the Cilk manual for more explanation\n"));
  WHEN_CILK_NDL(
  CILK_CHECK((char *) start_static_write + static_size <=
	      (char *) SHADOW_VAR(SHADOW_INFO(start_heap_write)),
	     ("Overlap between static writer info with heap writer info\n"
	      "Please change the shadow address and info address in cilk-nd.h\n"
	      "See the Cilk manual for more explanation\n"));
  CILK_CHECK((char *) SHADOW_INFO(start_heap_write) + heap_size <= 
	      (char *) SHADOW_VAR(SHADOW_INFO(start_static_write)),
	     ("Overlap between heap writer info with static writer info\n"
	      "Please change the shadow address and info address in cilk-nd.h\n"
	      "See the Cilk manual for more explanation\n"));
  )
  WHEN_SOLARIS(WHEN_CILK_NDL(
  CILK_CHECK((char *) SHADOW_INFO(start_static_write) + static_size <= 
	      (char *) USE_SHARED_RO(heap_arena_start),
	     ("Overlap between static writer info with heap\n"
	      "Please change the shadow address and info address in cilk-nd.h\n"
	      "See the Cilk manual for more explanation\n"));))

  /* now do mmap */
  p = mmap(start_static_read, static_size, PROT_READ | PROT_WRITE, 
	   MAP_SHARED | MAP_FIXED, static_reader_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for static_reader_fd\n"
	     "Please change the static reader shadow address STATIC_READER in cilk-nd.h\n"
	     "See the Cilk manual for more explanation\n");

  p = mmap(start_static_write, static_size, PROT_READ | PROT_WRITE, 
	   MAP_SHARED | MAP_FIXED, static_writer_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for static_writer_fd\n"
	     "Please change the static writer shadow address STATIC_WRITER in cilk-nd.h\n"
	     "See the Cilk manual for more explanation\n");

  p = mmap(start_heap_read, heap_size, PROT_READ | PROT_WRITE, 
	   MAP_SHARED | MAP_FIXED, heap_reader_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for heap_reader_fd\n"
	     "Please change the heap reader shadow address HEAP_READER in cilk-nd.h\n"
	     "See the Cilk manual for more explanation\n");

  p = mmap(start_heap_write, heap_size, PROT_READ | PROT_WRITE, 
	   MAP_SHARED | MAP_FIXED, heap_writer_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for heap_writer_fd\n"
	     "Please change the heap writer shadow address HEAP_WRITER in cilk-nd.h\n"
	     "See the Cilk manual for more explanation\n");

  WHEN_CILK_NDL(
  p = mmap((void *) SHADOW_VAR(SHADOW_INFO(start_static_read)), 
	   static_size + heap_size, PROT_READ | PROT_WRITE, 
	   MAP_SHARED | MAP_FIXED, static_reader_info_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for static_reader_info_fd\n"
	     "Please change the info address SHADOW_INFO and/or SHADOW_VAR in cilk-nd.h\n"
	     "See the Cilk manual for more explanation\n");

  p = mmap((void *) SHADOW_VAR(SHADOW_INFO(start_static_write)), 
	   static_size + heap_size, PROT_READ | PROT_WRITE, 
	   MAP_SHARED | MAP_FIXED, static_writer_info_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for static_writer_info_fd\n"
	     "Please change the info address SHADOW_INFO and/or SHADOW_VAR in cilk-nd.h\n"
	     "See the Cilk manual for more explanation\n");

  p = mmap((void *) SHADOW_VAR(SHADOW_INFO(start_heap_read)), 2*heap_size, 
	   PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, 
	   heap_reader_info_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for heap_reader_info_fd\n"
	     "Please change the info address SHADOW_INFO and/or SHADOW_VAR in cilk-nd.h\n"
	     "See the Cilk manual for more explanation\n");

  p = mmap((void *) SHADOW_VAR(SHADOW_INFO(start_heap_write)), 2*heap_size, 
	   PROT_READ | PROT_WRITE, MAP_SHARED | MAP_FIXED, 
	   heap_writer_info_fd, 0);
  if (p == MAP_FAILED)
    Cilk_die("mmap error in init_shadow_space for heap_writer_info_fd\n"
	     "Please change the info address SHADOW_INFO and/or SHADOW_VAR in cilk-nd.h\n"
	     "See the Cilk manual for more explanation\n");
  )
}

void nondeterminator_init()
{
  CILK_CHECK(USE_SHARED_RO(active_size) == 1,
	     ("The Nondeterminator runs only on one processor\n"));
  CILK_CHECK(PROC_ID_BIT + OPERATOR_BIT + TYPE_BIT <= 32,
	     ("Operator type for each shadow location need at least 6 bit"));

  init_private();
  init_disjoint_set();
  init_shadow_space();
  WHEN_CILK_ND_CACHE(init_cache());
}

WHEN_CILK_NDL(
static void save_shadow_info(DisjointSetMemberT *race_party)
{
  *(SHADOW_INFO(race_party)) = (USE_PRIVATE(race_linenum) << 8) +
                               find_filename(USE_PRIVATE(race_filename));
  *(SHADOW_VAR(SHADOW_INFO(race_party))) = USE_PRIVATE(race_name);
}

static inline char *strip_frame(char *name)
{
  if (strncmp(name, "_frame->scope", 13) == 0)
    return strchr(name, '.')+1;
  else
    return name;
}

static inline int find_filename(char *filename)
{
  FileListT *fl;
  int index;

  for (fl = USE_PRIVATE(filelist), index = 0; fl; fl = fl->next, index++) {
    if (filename == fl->filename)
      return fl->index;
  }
  fl = (FileListT *) malloc(sizeof(FileListT));
  fl->next = USE_PRIVATE(filelist);
  fl->filename = filename;
  fl->index = index;
  USE_PRIVATE(filelist) = fl;
  return index;
}

static inline char *get_filename(int index)
{
  FileListT *fl;

  for (fl = USE_PRIVATE(filelist); fl && (fl->index != index); fl = fl->next);
  CILK_CHECK(fl, ("error in retrieve filename in get_filename\n"));
  return fl->filename;
}
)

WHEN_CILK_ND_STATS(
void Cilk_spbags_start_timing()
{
  USE_PRIVATE(spbags_timing) = 1;
}

void Cilk_spbags_stop_timing()
{
  USE_PRIVATE(spbags_timing) = 0;
}

void Cilk_spbags_print_timing()
{
  Cilk_time time_now;

  time_now = gethrtime();

  USE_PRIVATE(read_access_time) = ADJUST_TIME(USE_PRIVATE(read_access_time), 
					      USE_PRIVATE(read_access_global_count));
  USE_PRIVATE(write_access_time) = ADJUST_TIME(USE_PRIVATE(write_access_time), 
					       USE_PRIVATE(write_access_global_count));
  USE_PRIVATE(init_time) = ADJUST_TIME(USE_PRIVATE(init_time), 
				       USE_PRIVATE(init_count));
  USE_PRIVATE(return_time) = ADJUST_TIME(USE_PRIVATE(return_time), 
					 USE_PRIVATE(return_count));
  USE_PRIVATE(sync_time) = ADJUST_TIME(USE_PRIVATE(sync_time), 
				       USE_PRIVATE(sync_count));
  
  fprintf(stderr, "time to profile read:  %f\n", 
    Cilk_wall_time_to_sec(USE_PRIVATE(read_access_time)));
  fprintf(stderr, "num of read access:    %lld\t(%f)\n", 
    USE_PRIVATE(read_access_count), 
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(read_access_time)) / 
	  USE_PRIVATE(read_access_count));
  fprintf(stderr, "num of global read:    %lld\t(%f)\n", 
    USE_PRIVATE(read_access_global_count), 
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(read_access_time)) / 
	  USE_PRIVATE(read_access_global_count));
  fprintf(stderr, "num of word read:      %lld\t(%f)\n\n", 
    USE_PRIVATE(read_access_word_count), 
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(read_access_time)) / 
	  USE_PRIVATE(read_access_word_count));

  fprintf(stderr, "time to profile write: %f\n", 
    Cilk_wall_time_to_sec(USE_PRIVATE(write_access_time)));
  fprintf(stderr, "num of write access:   %lld\t(%f)\n", 
    USE_PRIVATE(write_access_count), 
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(write_access_time)) / 
	  USE_PRIVATE(write_access_count));
  fprintf(stderr, "num of global write:   %lld\t(%f)\n", 
    USE_PRIVATE(write_access_global_count), 
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(write_access_time)) / 
	  USE_PRIVATE(write_access_global_count));
  fprintf(stderr, "num of word write:     %lld\t(%f)\n\n", 
    USE_PRIVATE(write_access_word_count), 
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(write_access_time)) /
	  USE_PRIVATE(write_access_word_count));

  fprintf(stderr, "time to init:          %f\n", 
    Cilk_wall_time_to_sec(USE_PRIVATE(init_time)));
  fprintf(stderr, "num of init:           %d\t(%f)\n\n", 
    USE_PRIVATE(init_count), 
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(init_time)) / USE_PRIVATE(init_count));

  fprintf(stderr, "time to return:        %f\n", 
    Cilk_wall_time_to_sec(USE_PRIVATE(return_time)));
  fprintf(stderr, "num of return_check:   %d\t(%f)\n\n", 
    USE_PRIVATE(return_count),
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(return_time))/USE_PRIVATE(return_count));

  fprintf(stderr, "time to sync:          %f\n", 
    Cilk_wall_time_to_sec(USE_PRIVATE(sync_time)));
  fprintf(stderr, "num of sync_check:     %d\t(%f)\n\n", 
    USE_PRIVATE(sync_count),
    WALL_TIME_TO_MICROSECONDS(USE_PRIVATE(sync_time))/USE_PRIVATE(sync_count));

  WHEN_CILK_ND_CACHE(
  fprintf(stderr, "num of read_hit_count: %lld\t(%f)\n", 
    USE_PRIVATE(read_access_cache_hit_count),
    ((float) USE_PRIVATE(read_access_cache_hit_count)) / 
	  USE_PRIVATE(read_access_count));
  fprintf(stderr, "num of write_hit_count:%lld\t(%f)\n\n", 
    USE_PRIVATE(write_access_cache_hit_count),
    ((float) USE_PRIVATE(write_access_cache_hit_count)) / 
	  USE_PRIVATE(write_access_count));
  fprintf(stderr, "num of read_hit_word_count: %lld\n", 
    USE_PRIVATE(read_access_cache_hit_word_count));
  fprintf(stderr, "num of write_hit_word_count:%lld\n\n", 
    USE_PRIVATE(write_access_cache_hit_word_count));
  fprintf(stderr, "local read access:     %lld\n",
    USE_PRIVATE(read_access_count) - USE_PRIVATE(read_access_global_count) - 
	  USE_PRIVATE(read_access_cache_hit_count));
  fprintf(stderr, "local write access:    %lld\n",
    USE_PRIVATE(write_access_count) - USE_PRIVATE(write_access_global_count) - 
	  USE_PRIVATE(write_access_cache_hit_count));
  )

  fprintf(stderr, "read/write overhead:   %f\n", 
    Cilk_wall_time_to_sec(USE_PRIVATE(read_access_time) + 
			  USE_PRIVATE(write_access_time)));
  fprintf(stderr, "other overhead:        %f\n", 
    Cilk_wall_time_to_sec(USE_PRIVATE(init_time) + USE_PRIVATE(return_time) + 
			  USE_PRIVATE(sync_time)));
  fprintf(stderr, "total overhead:        %f\n", 
    Cilk_wall_time_to_sec(USE_PRIVATE(read_access_time) + 
			  USE_PRIVATE(write_access_time) + 
			  USE_PRIVATE(init_time) + USE_PRIVATE(return_time) + 
			  USE_PRIVATE(sync_time)));
  fprintf(stderr, "proc id                %d\n",
    USE_PRIVATE(next_proc_id));
  fprintf(stderr, "timing overhead        %f\n",
    Cilk_wall_time_to_sec(((Cilk_time) (USE_PRIVATE(read_access_count)+
				       USE_PRIVATE(read_access_global_count)+
				       USE_PRIVATE(write_access_count)+
				       USE_PRIVATE(write_access_global_count)+
				       2*USE_PRIVATE(init_count)+
				       2*USE_PRIVATE(return_count)+
				       2*USE_PRIVATE(sync_count)))
			  * HRTIME + gethrtime() - time_now + HRTIME));
}     
)
#endif
