
/*
 * $Id: barrier.c,v 1.1.1.1.2.1 1997/03/04 03:10:53 fengmd Exp $
 */

/*
 * $Log: barrier.c,v $
 * Revision 1.1.1.1.2.1  1997/03/04 03:10:53  fengmd
 * Merge the change of the Cilk-5 base.
 *
 * Revision 1.2  1997/03/01 17:42:37  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/*
 * This file contains the code for barriers
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>

/***********************************************************\
 * Code for barriers
\***********************************************************/
DECLARE_SHARED_VARIABLE(static Cilk_mutex, barrier_lock, );
DECLARE_SHARED_VARIABLE(volatile int, barrier_counter, );
DECLARE_SHARED_VARIABLE(volatile int, barrier_release, );
DECLARE_PRIVATE_VARIABLE(int, barrier_direction, );

/* sense-reversing barrier, see Hennessy & Patterson page 702 */
/* this barrier can be safely reused */

/* 
 * Volker pointed out that the original code in H&P has a bug,
 * and provided a fix, which is used here
 */
void Cilk_barrier(void)
{
     int local_counter;

     USE_PRIVATE(barrier_direction) = !USE_PRIVATE(barrier_direction);

     Cilk_mutex_wait(&USE_SHARED(barrier_lock));
     local_counter = ++USE_SHARED(barrier_counter);
     Cilk_mutex_signal(&USE_SHARED(barrier_lock));

     if (local_counter == USE_SHARED_RO(active_size)) {
	  USE_SHARED(barrier_counter) = 0;
	  USE_SHARED(barrier_release) = USE_PRIVATE(barrier_direction);
     } else
	  while (USE_SHARED(barrier_release) != USE_PRIVATE(barrier_direction))
	       if (USE_SHARED_RO(yield_slice))
		    Cilk_yield();
}

void Cilk_barrier_init(void)
{
     /* initialize the barrier code */
     Cilk_mutex_init(&USE_SHARED(barrier_lock));
     USE_SHARED(barrier_counter) = 0;
     USE_PRIVATE(barrier_direction) = 0;
     USE_SHARED(barrier_release) = USE_PRIVATE(barrier_direction);
}
